#! /usr/bin/env python
# ______________________________________________________________________
# Module imports

import types

__DEBUG__ = False

# ______________________________________________________________________
# Class definitions

class MyContext (object):
    def __init__ (self, exp, env):
        self.exp = exp
        self.env = env

    def __repr__ (self):
        return "<%s(%r, %r)>" % (type(self).__name__, self.exp, self.env)

class MyCallContext (MyContext):
    pass

class MyTailCallContext (MyCallContext):
    pass

# ______________________________________________________________________

class MyResult (object):
    def __init__ (self, result):
        self.result = result

    def __repr__ (self):
        return 'MyResult(%r)' % (self.result,)

    def __str__ (self):
        return str(self.result)

# ______________________________________________________________________

class MyClosure (object):
    def __init__ (self, params, body, env):
        self.params = params
        self.body = body
        self.env = env

# ______________________________________________________________________

class MyInterp (object):
    def __init__ (self):
        self.step_count = 0
        self.max_call_depth = 0
        self.call_stack = []
        self.context_stack = []
        self.result_stack = []

    def push_context (self, context):
        if __DEBUG__:
            print "push_context(): => ", self.call_stack,
            print len(self.context_stack)
        # The following signals that we are entering a lambda body.
        if isinstance(context, MyCallContext):
            # The following signals that we are making a tail call,
            # and should to thow out all evaluation contexts up to the
            # last lambda body, and replace them with this one.
            if isinstance(context, MyTailCallContext):
                if len(self.call_stack):
                    del self.context_stack[self.call_stack.pop():]
                else:
                    self.context_stack.pop()
            self.call_stack.append(len(self.context_stack))
            if len(self.call_stack) > self.max_call_depth:
                self.max_call_depth = len(self.call_stack)
        self.context_stack.append(self.make_gen(context))
        if __DEBUG__:
            print "push_context(): <= ", self.call_stack,
            print len(self.context_stack)
            print

    def pop_context (self, result_obj):
        if __DEBUG__:
            print "pop_context(): =>", self.call_stack,
            print len(self.context_stack)
        if len(self.call_stack) > 0:
            if len(self.context_stack) < self.call_stack[-1]:
                self.call_stack.pop()
        self.context_stack.pop()
        assert len(self.result_stack) == 0
        self.result_stack.append(result_obj.result)
        if __DEBUG__:
            print "pop_context(): <=", self.call_stack,
            print len(self.context_stack)
            print

    def get_result (self):
        return self.result_stack.pop()

    def has_action (self):
        return len(self.context_stack) > 0

    def get_next_action (self):
        return self.context_stack[-1].next()

    def my_eval (self, exp, env):
        self.step_count = 0
        self.max_call_depth = 0
        self.push_context(MyContext(exp, env))
        while self.has_action():
            self.step_count += 1
            action = self.get_next_action()
            if __DEBUG__:
                print "my_eval(): action = %r" % (action,)
                print
            if isinstance(action, MyContext):
                self.push_context(action)
            elif isinstance(action, MyResult):
                self.pop_context(action)
        assert len(self.result_stack) == 1
        return self.get_result()

    def make_gen (self, context):
        exp, env = context.exp, context.env
        if __DEBUG__:
            print "make_gen():", exp, env
            print
        if type(exp) in (int, float, str, list, types.FunctionType,
                         types.NoneType):
            yield MyResult(exp)
        elif type(exp) == tuple:
            hd = exp[0]
            if hd == 'symbol':
                yield MyResult(env[exp[1]])
            elif hd == 'lambda':
                yield MyResult(MyClosure(exp[1], exp[2], env.copy()))
            elif hd == 'quote':
                children = []
                for child in exp[1:]:
                    yield MyContext(child, env)
                    children.append(self.get_result())
                yield MyResult(tuple(children))
            elif hd == 'app':
                yield MyContext(exp[1], env)
                fn = self.get_result()
                args = []
                for arg in exp[2:]:
                    yield MyContext(arg, env)
                    args.append(self.get_result())
                if type(fn) == types.FunctionType:
                    yield MyResult(fn(*args))
                else:
                    subenv = fn.env.copy()
                    subenv.update(dict(zip(fn.params, args)))
                    yield MyCallContext(fn.body, subenv)
                    yield MyResult(self.get_result())
            elif hd == 'tapp':
                # Tail call.  Pop the context, and push a replacement.
                yield MyContext(exp[1], env)
                fn = self.get_result()
                args = []
                for arg in exp[2:]:
                    yield MyContext(arg, env)
                    args.append(self.get_result())
                subenv = fn.env.copy()
                subenv.update(dict(zip(fn.params, args)))
                # Throw away the calling context, causing this
                # generator to go away, and not have to return a
                # result.
                yield MyTailCallContext(fn.body, subenv)
            elif hd == 'define':
                yield MyContext(exp[2], env)
                val = self.get_result()
                env[exp[1]] = val
                # XXX Hack to support back-patch of recursive definition
                if type(val) == MyClosure:
                    val.env[exp[1]] = val
                yield MyResult(val)
            elif hd == 'if':
                yield MyContext(exp[1], env)
                pred = self.get_result()
                if pred:
                    yield MyContext(exp[2], env)
                    yield MyResult(self.get_result())
                else:
                    yield MyContext(exp[3], env)
                    yield MyResult(self.get_result())
            elif hd == 'begin':
                crnt_result = None
                for crnt_exp in exp[1:]:
                    yield MyContext(crnt_exp, env)
                    crnt_result = self.get_result()
                yield MyResult(crnt_result)
        else:
            raise Exception("Unknown kind of expression: %r" % exp)

# ______________________________________________________________________
# Utility functions

def eq (val1, val2):
    return val1 == val2

def plus (val1, val2):
    return val1 + val2

def minus (val1, val2):
    return val1 - val2

def times (val1, val2):
    return val1 * val2

# ______________________________________________________________________
# Main (self-test) routine

test_progs = [
    ('begin',
     ('define', 'fac',
      ('lambda', ['x', 'acc'],
       ('if', ('app', eq, ('symbol', 'x'), 1),
        ('symbol', 'acc'),
        ('app', ('symbol', 'fac'),
         ('app', minus, ('symbol', 'x'), 1),
         ('app', times, ('symbol', 'x'), ('symbol', 'acc')))))),
     ('app', ('symbol', 'fac'), 200, 1)),
    ('begin',
     ('define', 'fac',
      ('lambda', ['x', 'acc'],
       ('if', ('app', eq, ('symbol', 'x'), 1),
        ('symbol', 'acc'),
        ('tapp', ('symbol', 'fac'), # Signal tail call here.
         ('app', minus, ('symbol', 'x'), 1),
         ('app', times, ('symbol', 'x'), ('symbol', 'acc')))))),
     ('tapp', ('symbol', 'fac'), 200, 1)),
    ]

if __name__ == "__main__":
    import sys
    if "-d" in sys.argv:
        __DEBUG__ = True
    interp = MyInterp()
    for test_prog in test_progs:
        print interp.my_eval(test_prog, {})
        print ("Computed in %d steps, with a max call depth of %d." %
               (interp.step_count, interp.max_call_depth))
        print

# ______________________________________________________________________
# End of interp_play.py
